#include <pthread.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

extern "C" {
#include <xmms/plugin.h>
#include <xmms/configfile.h>
#include <xmms/util.h>
}
#include <sidplay/player.h>
#include <gtk/gtk.h>

static void sid_init(void);
static void sid_about(void);
static void sid_configure(void);
static int sid_is_our_file(char *filename);
static void sid_play_file(char *filename);
static void sid_stop(void);
static void sid_pause(short paused);
static void sid_seek(int time);
static int sid_get_time(void);
static void sid_get_song_info(char *filename,char **title,int *length);
static void sid_file_info_box(char *filename);

InputPlugin sid_ip = {
  NULL,
  NULL,
  "SidPlay " VERSION,
  sid_init,
  sid_about,
  sid_configure,
  sid_is_our_file,
  NULL,
  sid_play_file,
  sid_stop,
  sid_pause,
  sid_seek,
  NULL,
  sid_get_time,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  sid_get_song_info,
  sid_file_info_box,
  NULL
};

emuConfig config = {
  44100,			// frequency
  SIDEMU_16BIT,			// bits per sample
  SIDEMU_SIGNED_PCM,		// sample fomat
  SIDEMU_STEREO,		// channels
  0,				// -
  SIDEMU_NONE,	 		// volume control
  true,				// mos8580
  true,				// measure volume
  true,				// emulate filter
  0.0,				// filter Fs
  0.0,				// filter Fm
  0.0,				// filter Ft
  MPU_PLAYSID_ENVIRONMENT,	// memory mode
  SIDTUNE_CLOCK_PAL,		// clock speed
  true,				// force song speed
  0,				// digi player scans
  SIDEMU_NONE			// auto panning
};

static gboolean use_STIL;
static gchar *STIL_path;

static pthread_t decode_thread;

static emuEngine *engine;
static sidTune *tune;

static bool going = false;

static int currentSong;

GtkWidget *confdialog;
GtkWidget *infodialog;
GtkWidget *infolabel;
GtkWidget *button8bit;
GtkWidget *button16bit;
GtkWidget *buttonMono;
GtkWidget *buttonStereo;
GtkWidget *menuFrequency;
GtkWidget *buttonSIDFilter;
GtkWidget *buttonMeasuredVolume;
GtkWidget *buttonMOSWaveforms;
GtkWidget *buttonForceSpeed;
GtkWidget *buttonFullBank;
GtkWidget *buttonTransparentRom;
GtkWidget *buttonPlaySIDEnv;
GtkWidget *buttonPAL;
GtkWidget *buttonNTSC;
GtkWidget *buttonUseSTIL;
GtkWidget *entrySTILpath;
GtkWidget *currentSongBar;

#define DEBUG(fmt, args...) printf(fmt "\n", ## args);

static void *play_loop(void *arg);

extern "C" InputPlugin *get_iplugin_info(void)
{
  return &sid_ip;
}

static int
change_current_song(int newCurrent)
{
  sidTuneInfo info;
  
  currentSong = newCurrent;

  if (currentSongBar) {
    gtk_adjustment_set_value(
      gtk_range_get_adjustment(GTK_RANGE(currentSongBar)), currentSong);
    gtk_range_slider_update(GTK_RANGE(currentSongBar));
  }

  if(going) {
    going = false;
    pthread_join(decode_thread, NULL);
    sid_ip.output->close_audio();
  }

  sidEmuInitializeSong(*engine, *tune, currentSong);

  tune->getInfo(info);
  if (info.songs == 0)
    return 0;

  sid_ip.set_info(g_strdup_printf("%s (%i/%i)", info.nameString, currentSong, info.songs), 
		  100000, 
		  config.channels * (config.bitsPerSample / 8) * config.frequency, 
		  config.frequency, config.channels);

  if (infolabel) {
    gtk_label_set_text(GTK_LABEL(infolabel), 
		       g_strdup_printf("%i/%i", currentSong, info.songs));
  }

  if(!sid_ip.output->
     open_audio((config.bitsPerSample == 16) ? FMT_S16_LE : FMT_U8,
		config.frequency,
		config.channels)) {
    return 0;
  }

  going = true;
  if(pthread_create(&decode_thread, NULL, play_loop, NULL) != 0) {
    return 0;
  }

  return 1;
}

static void 
sid_init(void)
{
  ConfigFile *cfgfile;
  gchar *filename;
  gboolean b;
  engine = new emuEngine();
  tune = new sidTune(0);

  STIL_path = g_strdup("");
  use_STIL = false;

  currentSongBar = NULL;

  filename = g_strconcat(g_get_home_dir(), "/.xmms/config", NULL);
  if (cfgfile = xmms_cfg_open_file(filename)) {
    xmms_cfg_read_int(cfgfile, "SidPlay", "frequency", (gint *)&config.frequency);
    xmms_cfg_read_int(cfgfile, "SidPlay", "bits_per_sample", (gint *)&config.bitsPerSample);
    xmms_cfg_read_int(cfgfile, "SidPlay", "channels", (gint *)&config.channels);
    if(xmms_cfg_read_boolean(cfgfile, "SidPlay", "emulate_filter", &b))
      config.emulateFilter = b;
    if(xmms_cfg_read_boolean(cfgfile, "SidPlay", "measured_volume", &b))
      config.measuredVolume = b;
    if(xmms_cfg_read_boolean(cfgfile, "SidPlay", "mos8580", &b))
      config.mos8580 = b;
    if(xmms_cfg_read_boolean(cfgfile, "SidPlay", "force_song_speed", &b))
      config.forceSongSpeed = b;
    if(xmms_cfg_read_boolean(cfgfile, "SidPlay", "use_stil", &b))
      use_STIL = b;
    xmms_cfg_read_string(cfgfile, "SidPlay", "stil_path", &STIL_path);
    xmms_cfg_read_int(cfgfile, "SidPlay", "memory_mode", (gint *)&config.memoryMode);
    xmms_cfg_read_int(cfgfile, "SidPlay", "clock_speed", (gint *)&config.clockSpeed);
    xmms_cfg_free(cfgfile);
  }

  g_free(filename);

  engine->setConfig(config);
  engine->setDefaultFilterStrength();
}

static void
sid_about(void)
{
  char text[256];
  static GtkWidget *dialog = NULL;
  GtkWidget *button;
  GtkWidget *vbox;
  GtkWidget *label;
  GtkWidget *bbox;

  if (!dialog) {
    dialog = gtk_dialog_new();
    gtk_window_set_wmclass(GTK_WINDOW(dialog), "about_dialog", "SidPlay");
    gtk_window_set_title(GTK_WINDOW(dialog), "About SidPlay plugin");
    gtk_window_set_policy(GTK_WINDOW(dialog), FALSE, FALSE, FALSE);
    gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_MOUSE);
    gtk_signal_connect(GTK_OBJECT(dialog), "destroy", 
		       GTK_SIGNAL_FUNC(gtk_widget_destroyed), &dialog);
  
    vbox = gtk_vbox_new(FALSE, 1);
    gtk_container_set_border_width(GTK_CONTAINER(vbox), 5);
    gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), vbox);

    sprintf(text, 
	    "SidPlay XMMS Input Plugin\n"
	    "by Fredrik Hallenberg <hallon@lysator.liu.se>\n\n"
	    "More info on http://www.lysator.liu.se/~hallon/sidplay.html\n\n"
	    "Using SidPlay %s by Michael Schwendt <sidplay@geocities.com>",
	    engine->getVersionString());
    label = gtk_label_new(text);
    gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, TRUE, 2);

    bbox = gtk_hbutton_box_new();
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), bbox, TRUE, TRUE, 5);
    gtk_button_box_set_child_size(GTK_BUTTON_BOX(bbox), 80, 0);
    gtk_button_box_set_spacing(GTK_BUTTON_BOX(bbox), 10);

    button = gtk_button_new_with_label("Ok");
    gtk_container_add(GTK_CONTAINER(bbox), button);
    gtk_signal_connect_object(GTK_OBJECT(button), "clicked", 
			      GTK_SIGNAL_FUNC(gtk_widget_destroy), 
			      GTK_OBJECT(dialog));
  }
  
  if (!GTK_WIDGET_VISIBLE(dialog))
    gtk_widget_show_all(dialog);
}

static void
configure(void)
{
  engine->setConfig(config);
}

static void
configure_freq_cb(GtkWidget *widget, gpointer data)
{
  config.frequency = atoi((char*)data);

  configure();
}

static void
configure_ok_cb(GtkWidget *widget, gpointer data)
{
  ConfigFile *cfgfile;
  gchar *filename;
  
  filename = g_strconcat(g_get_home_dir(), "/.xmms/config", NULL);
  if (!(cfgfile = xmms_cfg_open_file(filename)))
    cfgfile = xmms_cfg_new();

  xmms_cfg_write_int(cfgfile, "SidPlay", "frequency", config.frequency);
  xmms_cfg_write_int(cfgfile, "SidPlay", "bits_per_sample", config.bitsPerSample);
  xmms_cfg_write_int(cfgfile, "SidPlay", "channels", config.channels);
  xmms_cfg_write_boolean(cfgfile, "SidPlay", "emulate_filter", config.emulateFilter);
  xmms_cfg_write_boolean(cfgfile, "SidPlay", "measured_volume", config.measuredVolume);
  xmms_cfg_write_boolean(cfgfile, "SidPlay", "mos8580", config.mos8580);
  xmms_cfg_write_boolean(cfgfile, "SidPlay", "force_song_speed", config.forceSongSpeed);
  xmms_cfg_write_boolean(cfgfile, "SidPlay", "use_stil", use_STIL);
  xmms_cfg_write_string(cfgfile, "SidPlay", "stil_path", STIL_path);
  xmms_cfg_write_int(cfgfile, "SidPlay", "memory_mode", config.memoryMode);
  xmms_cfg_write_int(cfgfile, "SidPlay", "clock_speed", config.clockSpeed);
  xmms_cfg_write_file(cfgfile, filename);
  xmms_cfg_free(cfgfile);

  g_free(filename);

  gtk_widget_destroy(confdialog);
}

static void
configure_cb(GtkWidget *widget, gpointer data)
{
  config.bitsPerSample = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(button8bit)) ? 8 : 16;
  config.channels = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(buttonMono)) ? 1 : 2;
  config.emulateFilter = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(buttonSIDFilter)) ? 8 : 16;
  config.measuredVolume = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(buttonMeasuredVolume));
  config.mos8580 = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(buttonMOSWaveforms));
  config.forceSongSpeed = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(buttonForceSpeed));
  use_STIL = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(buttonUseSTIL));
  g_free(STIL_path);
  STIL_path = g_strdup(gtk_entry_get_text(GTK_ENTRY(entrySTILpath)));

  gtk_widget_set_sensitive(entrySTILpath, use_STIL);

  config.memoryMode = 
    gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(buttonFullBank)) ? 
    MPU_BANK_SWITCHING :
    gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(buttonTransparentRom)) ? 
    MPU_TRANSPARENT_ROM : MPU_PLAYSID_ENVIRONMENT;

  config.clockSpeed =
    gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(buttonPAL)) ? 
    SIDTUNE_CLOCK_PAL : SIDTUNE_CLOCK_NTSC;

  configure();
}

static void
sid_configure(void)
{
  GtkWidget *vbox;
  GtkWidget *bbox;
  GtkWidget *audiobox1, *audiobox2, *emubox1, *emubox2, *miscbox;
  GtkWidget *button;
  GtkWidget *frame;
  GtkWidget *menu;
  GtkWidget *label;
  GtkWidget *optionmenu;
  GtkWidget *item;
  GtkWidget *notebook;
  GtkWidget *entry;
  GSList *group;
  gchar* itemname[] = { 
    "8000", "11025", "16500", "22050", "27500", "32000", "37500", "44100",
    "48000"
  };
  int i;
  int history;

  confdialog = gtk_dialog_new();
  gtk_signal_connect(GTK_OBJECT(confdialog), "destroy",
		     GTK_SIGNAL_FUNC(gtk_widget_destroyed),
		     &confdialog);

  gtk_window_set_title(GTK_WINDOW(confdialog), "Configure SidPlay Plugin");
  gtk_window_set_wmclass(GTK_WINDOW(confdialog), "sidplay_configure", "xmms");

  notebook = gtk_notebook_new();
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(confdialog)->vbox), notebook, TRUE, TRUE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(notebook), 5);

  vbox = gtk_vbox_new(FALSE, 5);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), 5);
  label = gtk_label_new("Audio");
  gtk_notebook_append_page(GTK_NOTEBOOK(notebook), vbox, label);
  audiobox1 = gtk_hbox_new(FALSE, 5);
  gtk_box_pack_start(GTK_BOX(vbox), audiobox1, FALSE, FALSE, 0);
  audiobox2 = gtk_hbox_new(FALSE, 5);
  gtk_box_pack_start(GTK_BOX(vbox), audiobox2, FALSE, FALSE, 0);
  
  vbox = gtk_vbox_new(FALSE, 5);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), 5);
  label = gtk_label_new("Emulator");
  gtk_notebook_append_page(GTK_NOTEBOOK(notebook), vbox, label);
  emubox1 = gtk_hbox_new(FALSE, 5);
  gtk_box_pack_start(GTK_BOX(vbox), emubox1, FALSE, FALSE, 0);
  emubox2 = gtk_hbox_new(FALSE, 5);
  gtk_box_pack_start(GTK_BOX(vbox), emubox2, FALSE, FALSE, 0);

  miscbox = gtk_vbox_new(FALSE, 5);
  gtk_container_set_border_width(GTK_CONTAINER(miscbox), 5);
  label = gtk_label_new("Misc");
  gtk_notebook_append_page(GTK_NOTEBOOK(notebook), miscbox, label);

  /*** STIL ***/

  frame = gtk_frame_new("STIL");
  gtk_box_pack_start(GTK_BOX(miscbox), frame, FALSE, FALSE, 0);

  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), 5);
  gtk_container_add(GTK_CONTAINER(frame), vbox);

  button = gtk_check_button_new_with_label("Use STIL");
  gtk_box_pack_start(GTK_BOX(vbox), button, FALSE, FALSE, 0);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), use_STIL);
  buttonUseSTIL = button;

  label = gtk_label_new("STIL path");
  gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);

  entry = gtk_entry_new();
  gtk_entry_set_text(GTK_ENTRY(entry), STIL_path);
  gtk_box_pack_start(GTK_BOX(vbox), entry, FALSE, FALSE, 0);
  entrySTILpath = entry;

  gtk_signal_connect(GTK_OBJECT(buttonUseSTIL), "clicked",
		     GTK_SIGNAL_FUNC(configure_cb), NULL);
  gtk_signal_connect(GTK_OBJECT(entrySTILpath), "focus_out_event",
		     GTK_SIGNAL_FUNC(configure_cb), NULL);

  /*** Resolution ***/

  frame = gtk_frame_new("Sample Resolution");
  gtk_box_pack_start(GTK_BOX(audiobox1), frame, TRUE, TRUE, 0);

  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), 5);
  gtk_container_add(GTK_CONTAINER(frame), vbox);
    
  button = gtk_radio_button_new_with_label(NULL, "8 bit");
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), 
			       config.bitsPerSample == SIDEMU_8BIT);
  gtk_box_pack_start(GTK_BOX(vbox), button, FALSE, FALSE, 0);
  button8bit = button;

  button = gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(button),
						       "16 bit");
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), 
			       config.bitsPerSample == SIDEMU_16BIT);
  gtk_box_pack_start(GTK_BOX(vbox), button, FALSE, FALSE, 0);
  button16bit = button;

  gtk_signal_connect(GTK_OBJECT(button8bit), "clicked",
		     GTK_SIGNAL_FUNC(configure_cb), NULL);
  gtk_signal_connect(GTK_OBJECT(button16bit), "clicked",
		     GTK_SIGNAL_FUNC(configure_cb), NULL);

  /*** Channels ***/

  frame = gtk_frame_new("Channels");
  gtk_box_pack_start(GTK_BOX(audiobox2), frame, TRUE, TRUE, 0);

  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), 5);
  gtk_container_add(GTK_CONTAINER(frame), vbox);
    
  button = gtk_radio_button_new_with_label(NULL, "Mono");
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), 
			       config.channels == SIDEMU_MONO);
  gtk_box_pack_start(GTK_BOX(vbox), button, FALSE, FALSE, 0);
  buttonMono = button;

  button = gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(button),
						       "Stereo");
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), 
			       config.channels == SIDEMU_STEREO);
  gtk_box_pack_start(GTK_BOX(vbox), button, FALSE, FALSE, 0);
  buttonStereo = button;

  gtk_signal_connect(GTK_OBJECT(buttonMono), "clicked",
		     GTK_SIGNAL_FUNC(configure_cb), NULL);
  gtk_signal_connect(GTK_OBJECT(buttonStereo), "clicked",
		     GTK_SIGNAL_FUNC(configure_cb), NULL);

  /*** Frequency ***/

  frame = gtk_frame_new("Frequency");
  gtk_box_pack_start(GTK_BOX(audiobox1), frame, TRUE, TRUE, 0);

  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), 5);
  gtk_container_add(GTK_CONTAINER(frame), vbox);

  menu = gtk_menu_new();
  menuFrequency = menu;

  group = NULL;
  history = 0;
  for(i = 0; i < 9; i++) {
    item = gtk_radio_menu_item_new_with_label(group, itemname[i]);
    gtk_signal_connect(GTK_OBJECT(item), "activate",
		       GTK_SIGNAL_FUNC(configure_freq_cb), itemname[i]);
    if (atoi(itemname[i]) == config.frequency) {
      history = i;
    }
    group = gtk_radio_menu_item_group(GTK_RADIO_MENU_ITEM(item));
    gtk_menu_append(GTK_MENU(menu), item);
  }

  optionmenu = gtk_option_menu_new();
  gtk_option_menu_set_menu(GTK_OPTION_MENU(optionmenu), menu);
  gtk_option_menu_set_history(GTK_OPTION_MENU(optionmenu), history);
  gtk_box_pack_start(GTK_BOX(vbox), optionmenu, FALSE, FALSE, 0);

  /*** General Settings ***/

  frame = gtk_frame_new("General Settings");
  gtk_box_pack_start(GTK_BOX(emubox1), frame, TRUE, TRUE, 0);

  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), 0);
  gtk_container_add(GTK_CONTAINER(frame), vbox);
    
  button = gtk_check_button_new_with_label("SID filter");
  gtk_box_pack_start(GTK_BOX(vbox), button, FALSE, FALSE, 0);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), config.emulateFilter);
  buttonSIDFilter = button;

  button = gtk_check_button_new_with_label("Measured volume");
  gtk_box_pack_start(GTK_BOX(vbox), button, FALSE, FALSE, 0);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), config.measuredVolume);
  buttonMeasuredVolume = button;

  button = gtk_check_button_new_with_label("MOS 8580 waveforms");
  gtk_box_pack_start(GTK_BOX(vbox), button, FALSE, FALSE, 0);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), config.mos8580);
  buttonMOSWaveforms = button;

  button = gtk_check_button_new_with_label("Force song speed");
  gtk_box_pack_start(GTK_BOX(vbox), button, FALSE, FALSE, 0);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), config.forceSongSpeed);
  buttonForceSpeed = button;

  gtk_signal_connect(GTK_OBJECT(buttonSIDFilter), "clicked",
		     GTK_SIGNAL_FUNC(configure_cb), NULL);
  gtk_signal_connect(GTK_OBJECT(buttonMeasuredVolume), "clicked",
		     GTK_SIGNAL_FUNC(configure_cb), NULL);
  gtk_signal_connect(GTK_OBJECT(buttonMOSWaveforms), "clicked",
		     GTK_SIGNAL_FUNC(configure_cb), NULL);
  gtk_signal_connect(GTK_OBJECT(buttonForceSpeed), "clicked",
		     GTK_SIGNAL_FUNC(configure_cb), NULL);

  /*** MPU memory mode ***/

  frame = gtk_frame_new("MPU memory mode");
  gtk_box_pack_start(GTK_BOX(emubox2), frame, TRUE, TRUE, 0);

  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), 0);
  gtk_container_add(GTK_CONTAINER(frame), vbox);

  button = gtk_radio_button_new_with_label(NULL, "Full bank-switching");
  gtk_box_pack_start(GTK_BOX(vbox), button, FALSE, FALSE, 0);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button),
			       config.memoryMode == MPU_BANK_SWITCHING);
  buttonFullBank = button;

  button = gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(button),
						       "Transparent ROM");
  gtk_box_pack_start(GTK_BOX(vbox), button, FALSE, FALSE, 0);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button),
			       config.memoryMode == MPU_TRANSPARENT_ROM);
  buttonTransparentRom = button;

  button = gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(button),
						       "PlaySID environment");
  gtk_box_pack_start(GTK_BOX(vbox), button, FALSE, FALSE, 0);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button),
			       config.memoryMode == MPU_PLAYSID_ENVIRONMENT);
  buttonPlaySIDEnv = button;

  gtk_signal_connect(GTK_OBJECT(buttonFullBank), "clicked",
		     GTK_SIGNAL_FUNC(configure_cb), NULL);
  gtk_signal_connect(GTK_OBJECT(buttonTransparentRom), "clicked",
		     GTK_SIGNAL_FUNC(configure_cb), NULL);
  gtk_signal_connect(GTK_OBJECT(buttonPlaySIDEnv), "clicked",
		     GTK_SIGNAL_FUNC(configure_cb), NULL);

  /*** C64 clock speed ***/

  frame = gtk_frame_new("C64 clock speed");
  gtk_box_pack_start(GTK_BOX(emubox1), frame, TRUE, TRUE, 0);

  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), 0);
  gtk_container_add(GTK_CONTAINER(frame), vbox);

  button = gtk_radio_button_new_with_label(NULL, "PAL");
  gtk_box_pack_start(GTK_BOX(vbox), button, FALSE, FALSE, 0);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), 
			       config.clockSpeed == SIDTUNE_CLOCK_PAL);
  buttonPAL = button;

  button = gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(button), "NTSC");
  gtk_box_pack_start(GTK_BOX(vbox), button, FALSE, FALSE, 0);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), 
			       config.clockSpeed == SIDTUNE_CLOCK_NTSC);
  buttonNTSC = button;

  gtk_signal_connect(GTK_OBJECT(buttonPAL), "clicked",
		     GTK_SIGNAL_FUNC(configure_cb), NULL);
  gtk_signal_connect(GTK_OBJECT(buttonNTSC), "clicked",
		     GTK_SIGNAL_FUNC(configure_cb), NULL);

  /*** Action Area ***/

  bbox = gtk_hbutton_box_new();
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(confdialog)->action_area), bbox, TRUE, TRUE, 5);
  gtk_button_box_set_child_size(GTK_BUTTON_BOX(bbox), 80, 0);
  gtk_button_box_set_spacing(GTK_BUTTON_BOX(bbox), 10);
  gtk_button_box_set_layout(GTK_BUTTON_BOX(bbox), GTK_BUTTONBOX_END);
    
  button = gtk_button_new_with_label("Ok");
  gtk_signal_connect_object(GTK_OBJECT(button), "clicked",
			    GTK_SIGNAL_FUNC(configure_ok_cb),
			    GTK_OBJECT(confdialog));
  gtk_container_add(GTK_CONTAINER(bbox), button);
  GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
  gtk_widget_grab_default(button);

  button = gtk_button_new_with_label("Cancel");
  gtk_signal_connect_object(GTK_OBJECT(button), "clicked",
			    GTK_SIGNAL_FUNC(gtk_widget_destroy),
			    GTK_OBJECT(confdialog));
  gtk_container_add(GTK_CONTAINER(bbox), button);
  GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
  
  gtk_widget_show_all(confdialog);
}

static int 
sid_is_our_file(char *filename)
{
  char *ext;
  ext = strrchr(filename, '.');
  if (ext && strcasecmp(ext, ".sid") == 0) {
    return 1;
  }
  return 0;
}

static void *
play_loop(void *arg)
{
  int length;
  char buffer[2048];

  while(going) {
    length = sizeof(buffer);

    sidEmuFillBuffer(*engine, *tune, buffer, length);

    sid_ip.add_vis_pcm(sid_ip.output->written_time(),
		       (config.bitsPerSample == 16) ? FMT_S16_LE : FMT_U8,
		       config.channels, length, buffer);

    if(effects_enabled()) {
      length = get_current_effect_plugin()->
	mod_samples((gpointer *)buffer, length, 
		    (config.bitsPerSample == 16) ? FMT_S16_LE : FMT_U8, 
		    config.frequency, config.channels);
    } 

    while (sid_ip.output->buffer_free() < length && going)
      xmms_usleep(10000);
    if (going)
      sid_ip.output->write_audio(buffer, length);
  }

  pthread_exit(NULL);
}

static gboolean
strequal(gchar* a, gchar* b)
{
  while (*a != '\0') {
    if (*a++ != *b++) {
      return false;
    }
  }
  return true;
}

static void
STIL(char *filename)
{
  FILE* stil;
  int len = strlen(filename);
  char buf[256];
  bool go;
  static GtkWidget *dialog = NULL;
  static GtkWidget *text;
  GtkWidget *button;
  GtkWidget *hbox;
  GtkWidget *scrollbar;
  GtkWidget *bbox;

  if (!dialog) {
    dialog = gtk_dialog_new();
    gtk_window_set_wmclass(GTK_WINDOW(dialog), "stil_dialog", "SidPlay");
    gtk_window_set_title(GTK_WINDOW(dialog), "STIL");
    gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_MOUSE);
    gtk_widget_set_usize(dialog, 520, 200);
    gtk_signal_connect(GTK_OBJECT(dialog), "destroy", 
		       GTK_SIGNAL_FUNC(gtk_widget_destroyed), &dialog);
  
    hbox = gtk_hbox_new(FALSE, 5);
    gtk_container_set_border_width(GTK_CONTAINER(hbox), 5);
    gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), hbox);

    text = gtk_text_new(NULL, NULL);
    gtk_text_set_editable(GTK_TEXT(text), FALSE);
    gtk_box_pack_start(GTK_BOX(hbox), text, TRUE, TRUE, 0);

    scrollbar = gtk_vscrollbar_new(GTK_TEXT(text)->vadj);
    gtk_box_pack_start(GTK_BOX(hbox), scrollbar, FALSE, FALSE, 0);

    bbox = gtk_hbutton_box_new();
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), bbox, TRUE, TRUE, 5);
    gtk_button_box_set_child_size(GTK_BUTTON_BOX(bbox), 80, 0);
    gtk_button_box_set_spacing(GTK_BUTTON_BOX(bbox), 10);

    button = gtk_button_new_with_label("Ok");
    gtk_container_add(GTK_CONTAINER(bbox), button);
    gtk_signal_connect_object(GTK_OBJECT(button), "clicked", 
			      GTK_SIGNAL_FUNC(gtk_widget_destroy), 
			      GTK_OBJECT(dialog));
  }

  if ((stil = fopen(STIL_path, "r")) == NULL) {
    g_message("Can't open STIL file");
    return;
  }

  gtk_text_backward_delete(GTK_TEXT(text),
			   gtk_text_get_length(GTK_TEXT(text)));
  
  go = false;
  while (!feof(stil)) {
    fgets(buf, sizeof(buf), stil);
    g_strchomp(buf);
    if (buf[0] == '/') {
      if (go == true) {
	break;
      }
      else if(strequal(buf, &filename[len - strlen(buf)])) {
	go = true;
      }
    }
    else if (go) {
      gtk_text_insert(GTK_TEXT(text), NULL, NULL, NULL,
		      buf, strlen(buf));
      gtk_text_insert(GTK_TEXT(text), NULL, NULL, NULL,
		      "\n", 1);
    }
  }
  fclose(stil);

  if (go && !GTK_WIDGET_VISIBLE(dialog))
    gtk_widget_show_all(dialog);
  if (!go)
    gtk_widget_destroy(dialog);
}

static void
sid_play_file(char *filename)
{
  sidTuneInfo info;

  if(!tune->open(filename)) {
    return;
  }

  if (use_STIL && STIL_path[0])
    STIL(filename);

  if (infodialog)
    sid_file_info_box(filename);

  tune->getInfo(info);
  change_current_song(info.startSong);
}

static void
sid_stop(void)
{
  if(going) {
    going = false;
    pthread_join(decode_thread, NULL);
    sid_ip.output->close_audio();
  }
}

static void
sid_pause(short paused)
{
  sid_ip.output->pause(paused);
}

static void
sid_seek(int time)
{
  sidTuneInfo info;
  tune->getInfo(info);

  if (time == 0 && currentSong > 1) {
    change_current_song(currentSong - 1);
  }
  else if (time == 100 && currentSong < info.songs) {
    change_current_song(currentSong + 1);
  }
}

static int 
sid_get_time(void)
{
  if (!going)
    return -1;
  else
    return sid_ip.output->output_time();
}

static void
sid_get_song_info(char *filename,char **title,int *length)
{
  sidTune ttune(filename);
  sidTuneInfo info;

  ttune.getInfo(info);
  *title = g_strdup_printf("%s (%i/%i)", info.nameString, info.startSong, info.songs);

  *length = 0;
}

static void
tune_changed_cb(GtkWidget *widget, gpointer data)
{
  change_current_song((int)GTK_ADJUSTMENT(widget)->value);
}

static void
info_box_closed(GtkWidget *widget, gpointer data)
{
  gtk_widget_destroy(infodialog);
  infodialog = NULL;
  infolabel = NULL;
  currentSongBar = NULL;
}

static void
sid_file_info_box(char *filename)
{
  gchar text[1024];
  gchar buf[256];
  sidTune ttune(filename);
  sidTuneInfo info;
  int i;
  GtkWidget *button;
  GtkWidget *vbox;
  GtkWidget *label;
  GtkObject *adjustment;

  if (infodialog) {
    gtk_widget_destroy(infodialog);
  }

  infodialog = gtk_dialog_new();
  gtk_window_set_wmclass(GTK_WINDOW(infodialog), "info_dialog", "SidPlay");
  gtk_window_set_title(GTK_WINDOW(infodialog), "File Info");
  gtk_window_set_policy(GTK_WINDOW(infodialog), FALSE, FALSE, FALSE);
  gtk_window_set_position(GTK_WINDOW(infodialog), GTK_WIN_POS_MOUSE);
  gtk_widget_set_usize(infodialog, 200, 150);
  gtk_signal_connect(GTK_OBJECT(infodialog), "destroy", 
		     GTK_SIGNAL_FUNC(info_box_closed), NULL);
  
  vbox = gtk_vbox_new(FALSE, 2);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), 5);
  gtk_container_add(GTK_CONTAINER(GTK_DIALOG(infodialog)->vbox), vbox);

  ttune.getInfo(info);
  text[0] = '\0';
  for(i = 0; i < info.numberOfInfoStrings; i++) {
    strcat(text, info.infoString[i]);
    strcat(text, "\n");
  }  
  text[strlen(text) - 1] = '\0';
  label = gtk_label_new(text);
  gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, TRUE, 2);

  infolabel = gtk_label_new(g_strdup_printf("%i/%i", currentSong, info.songs));
  gtk_box_pack_start(GTK_BOX(vbox), infolabel, FALSE, TRUE, 2);

  adjustment = gtk_adjustment_new(currentSong, 1.0, info.songs + 1, 
				  1.0, 1.0, 1.0);
  currentSongBar = gtk_hscrollbar_new(GTK_ADJUSTMENT(adjustment));
  gtk_range_set_update_policy(GTK_RANGE(currentSongBar), GTK_UPDATE_CONTINUOUS);
  gtk_signal_connect(GTK_OBJECT(adjustment), "value_changed",
		     GTK_SIGNAL_FUNC(tune_changed_cb), NULL);
  gtk_box_pack_start(GTK_BOX(vbox), currentSongBar, TRUE, FALSE, 2);

  button = gtk_button_new_with_label("Ok");
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(infodialog)->action_area), button, 
		     TRUE, TRUE, 5);
  gtk_signal_connect_object(GTK_OBJECT(button), "clicked", 
			    GTK_SIGNAL_FUNC(info_box_closed), NULL);

  gtk_widget_show_all(infodialog);
}
